from .checkpoint import (
    CheckpointManager as CheckpointManager,
    load_and_convert_checkpoint_to_torch as load_and_convert_checkpoint_to_torch,
    float_tensor_to_dtype as float_tensor_to_dtype,
    read_ckpt as read_ckpt,
    save_ckpt as save_ckpt,
    get_dtype as get_dtype,
)

from .partition_utils import (
    get_jax_mesh as get_jax_mesh,
    names_in_current_mesh as names_in_current_mesh,
    get_names_from_partition_spec as get_names_from_partition_spec,
    match_partition_rules as match_partition_rules,
    flatten_tree as flatten_tree,
    get_metrics as get_metrics,
    tree_apply as tree_apply,
    get_weight_decay_mask as get_weight_decay_mask,
    named_tree_map as named_tree_map,
    tree_path_to_string as tree_path_to_string,
    make_shard_and_gather_fns as make_shard_and_gather_fns,
    with_sharding_constraint as with_sharding_constraint,
    wrap_function_with_rng as wrap_function_with_rng,
    create_mesh as create_mesh
)

from .monitor import (
    run as smi_run,
    get_memory_information as get_memory_information,
    is_notebook as is_notebook,
    threaded_log as smi_threaded_log,
    initialise_tracking as smi_initialise_tracking
)

from .func import (
    average_metrics as average_metrics,
    global_norm as global_norm,
    transpose as transpose,
    auxiliary_load_balancing_loss_func as auxiliary_load_balancing_loss_func,
    get_loss_normalizing_factor_and_weights as get_loss_normalizing_factor_and_weights,
    convert_special_loss_normalizing_factor_to_enum as convert_special_loss_normalizing_factor_to_enum,
    SpecialLossNormalizingFactor as SpecialLossNormalizingFactor,
    cross_entropy_loss_and_accuracy as cross_entropy_loss_and_accuracy,
    fused_cross_entropy_loss_and_accuracy as fused_cross_entropy_loss_and_accuracy,
    compute_weighted_cross_entropy_and_accuracy as compute_weighted_cross_entropy_and_accuracy,
    compute_weighted_cross_entropy as compute_weighted_cross_entropy,
    binary_cross_entropy_onehot as binary_cross_entropy_onehot,
    binary_cross_entropy as binary_cross_entropy,
    cross_entropy as cross_entropy,
    cross_entropy_with_logits as cross_entropy_with_logits,
    cross_entropy_onehot as cross_entropy_onehot,
    l2 as l2,
    hinge as hinge,
    mae as mae,
    nll as nll,
    mse_loss as mse_loss,
    mse as mse,
)

from .utils import (
    JaxRNG as JaxRNG,
    GenerateRNG as GenerateRNG,
    init_rng as init_rng,
    next_rng as next_rng,
)

from . import pallas_operations as pallas_operations
from . import optimizers as optimizers
from . import linen as linen

__version__ = "0.0.58"

__all__ = (
    # Loss and extra function

    "transpose",
    "global_norm",
    "average_metrics",
    "auxiliary_load_balancing_loss_func",
    "get_loss_normalizing_factor_and_weights",
    "convert_special_loss_normalizing_factor_to_enum",
    "SpecialLossNormalizingFactor",
    "mse_loss",
    "mse",
    "cross_entropy_loss_and_accuracy",
    "fused_cross_entropy_loss_and_accuracy",
    "compute_weighted_cross_entropy_and_accuracy",
    "compute_weighted_cross_entropy",
    "binary_cross_entropy_onehot",
    "binary_cross_entropy",
    "cross_entropy",
    "cross_entropy_with_logits",
    "cross_entropy_onehot",
    "l2",
    "hinge",
    "mae",
    "nll",

    # RNG Utils

    "JaxRNG",
    "GenerateRNG",
    "init_rng",
    "next_rng",

    # Monitor Functions

    "smi_run",
    "get_memory_information",
    "is_notebook",
    "smi_threaded_log",
    "smi_initialise_tracking",

    # Partition Utils
    "get_jax_mesh",
    "names_in_current_mesh",
    "get_names_from_partition_spec",
    "match_partition_rules",
    "flatten_tree",
    "get_metrics",
    "tree_apply",
    "get_weight_decay_mask",
    "named_tree_map",
    "tree_path_to_string",
    "make_shard_and_gather_fns",
    "with_sharding_constraint",
    "wrap_function_with_rng",
    "create_mesh",

    # Checkpointing Utils
    "CheckpointManager",
    "load_and_convert_checkpoint_to_torch",
    "float_tensor_to_dtype",
    "read_ckpt",
    "save_ckpt",
    "get_dtype",

    # Pallas Operations
    "pallas_operations",

    # Optimizers
    "optimizers",

    # Linen
    "linen"
)



import jax.numpy as jnp
from fjformer.func.loss_func import (
    compute_weighted_cross_entropy_and_accuracy,
    cross_entropy_loss_and_accuracy,
    cross_entropy_with_logits,
    get_loss_normalizing_factor_and_weights,
    fused_cross_entropy_loss_and_accuracy,
    SpecialLossNormalizingFactor,
)
import numpy as np
from optax import softmax_cross_entropy


def test_compute_weighted_cross_entropy_1():
    """
    batch_size = 2
    length = 4
    vocab_size = 8

    :return:
    """
    label_smoothing = 0.0
    z_loss_coeff = 0.0001

    logits = jnp.array(
        [
            [
                [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8],
                [0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1],
                [0.2, 0.3, 0.5, 0.7, 0.11, 0.13, 0.17, 0.19],
                [0.19, 0.17, 0.13, 0.11, 0.7, 0.5, 0.3, 0.2],
            ],
            [
                [0.5, 0.4, 0.3, 0.2, 0.1, 0.8, 0.7, 0.6],
                [0.6, 0.5, 0.4, 0.3, 0.2, 0.1, 0.8, 0.7],
                [0.3, 0.5, 0.7, 0.11, 0.13, 0.17, 0.19, 0.2],
                [0.2, 0.19, 0.17, 0.13, 0.11, 0.7, 0.5, 0.3],
            ],
        ],
        dtype=jnp.float32,
    )
    targets = jnp.array(
        [
            [7, 1, 3, 3],  # have two correct predictions based on highest logits
            [5, 6, 2, 7],  # have two correct predictions
        ],
        dtype=jnp.int32,
    )
    decoder_loss_weights = jnp.array([[1, 1, 1, 1], [1, 1, 1, 1]], dtype=jnp.float32)

    batch = {
        "decoder_target_tokens": targets,
        "decoder_loss_weights": decoder_loss_weights,
    }
    loss_normalizing_factor = SpecialLossNormalizingFactor.NUM_REAL_TARGET_TOKENS
    lnf, weights = get_loss_normalizing_factor_and_weights(
        loss_normalizing_factor, batch
    )

    (
        wce_loss,
        wce_z_loss,
        wce_weight_sum,
        wce_accuracy,
    ) = compute_weighted_cross_entropy_and_accuracy(
        logits=logits,
        targets=targets,
        weights=weights,
        label_smoothing=label_smoothing,
        z_loss=z_loss_coeff,
        loss_normalizing_factor=lnf,
    )

    print(
        f"Loss: {wce_loss}, Z-Loss: {wce_z_loss}, Weight Sum: {wce_weight_sum}, Accuracy: {wce_accuracy}"
    )
    ce_loss, ce_accuracy = cross_entropy_loss_and_accuracy(logits, targets)
    print(f"Cross Entropy Loss: {ce_loss}, Accuracy: {wce_accuracy}")

    assert np.isclose(ce_loss, wce_loss, rtol=1e-3, atol=1e-3), "Cross Entropy Losses do not match"
    assert np.isclose(ce_accuracy, wce_accuracy, rtol=1e-2, atol=1e-2), "Accuracies do not match"


def test_compute_weighted_cross_entropy_ignore_token_minus100():
    """
    batch_size = 2
    length = 4
    vocab_size = 8

    :return:
    """
    label_smoothing = 0.0
    z_loss_coeff = 0.0000

    logits = jnp.array(
        [
            [
                [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8],
                [0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1],
                [0.2, 0.3, 0.5, 0.7, 0.11, 0.13, 0.17, 0.19],
                [0.19, 0.17, 0.13, 0.11, 0.7, 0.5, 0.3, 0.2],
            ],
            [
                [0.5, 0.4, 0.3, 0.2, 0.1, 0.8, 0.7, 0.6],
                [0.6, 0.5, 0.4, 0.3, 0.2, 0.1, 0.8, 0.7],
                [0.3, 0.5, 0.7, 0.11, 0.13, 0.17, 0.19, 0.2],
                [0.2, 0.19, 0.17, 0.13, 0.11, 0.7, 0.5, 0.3],
            ],
        ],
        dtype=jnp.float32,
    )
    targets = jnp.array(
        [
            [-100, 1, 3, 3],  # Ignore first token
            [-100, -100, 2, 7],  # Ignore first two tokens
        ],
        dtype=jnp.int32,
    )
    decoder_loss_weights = (
        None  # Will be calculated by get_loss_normalizing_factor_and_weights
    )

    batch = {
        "decoder_target_tokens": targets,
        "decoder_loss_weights": decoder_loss_weights,
    }
    loss_normalizing_factor = SpecialLossNormalizingFactor.NUM_REAL_TARGET_TOKENS
    lnf, weights = get_loss_normalizing_factor_and_weights(
        loss_normalizing_factor, batch
    )

    (
        wce_loss,
        wce_z_loss,
        wce_weight_sum,
        wce_accuracy,
    ) = compute_weighted_cross_entropy_and_accuracy(
        logits=logits,
        targets=targets,
        weights=weights,
        label_smoothing=label_smoothing,
        z_loss=z_loss_coeff,
        loss_normalizing_factor=lnf,
    )

    print(
        f"Loss: {wce_loss}, Z-Loss: {wce_z_loss}, Weight Sum: {wce_weight_sum}, Accuracy: {wce_accuracy}"
    )

    # Construct valid array based on targets
    valid = jnp.where(targets > 0, 1, 0)
    ce_loss, ce_accuracy = cross_entropy_loss_and_accuracy(logits, targets, valid)
    print(f"Cross Entropy Loss: {ce_loss}, Accuracy: {ce_accuracy}")

    assert np.isclose(ce_loss, wce_loss, rtol=1e-2, atol=1e-2), "Cross Entropy Losses do not match"
    assert np.isclose(ce_accuracy, wce_accuracy, rtol=1e-1, atol=1e-1), "Accuracies do not match"


def test_multiple():
    # Logits: Unnormalized log probabilities from a model, shaped as [batch_size, num_classes].
    # Here, batch_size=2 for two examples, and num_classes=3 for three possible classes.
    logits = jnp.array(
        [
            [2.0, 1.0, 0.1],  # Logits for the first example
            [0.1, 2.0, 1.0],  # Logits for the second example
        ]
    )

    # Labels: One-hot encoded true class labels, with the same shape as logits [batch_size, num_classes].
    # Each row is a one-hot vector representing the true class of each example.
    labels = jnp.array(
        [
            [1, 0, 0],  # True class is the first one for the first example
            [0, 1, 0],  # True class is the second one for the second example
        ]
    )

    # Convert one-hot encoded labels to class indices (tokens)
    # Tokens are class indices representing the true classes for each example in the batch.
    # Shaped as [batch_size], where each entry is the index (0-based) of the true class.
    tokens = jnp.argmax(labels, axis=1)

    # Create a valid array  (a.k.a. decoder_loss_weights) that matches the shape of logits
    # Here, we're assuming all positions in all sequences are valid
    valid = jnp.ones_like(
        tokens, dtype=jnp.float32
    )  # Shape now correctly matches logits: (2, 3)
    print(f"Tokens: {tokens}, Valid: {valid}, valid_shape: {valid.shape}")

    # Computation using logits and class indices (tokens) as the representation of true labels.
    ce_loss, ce_accuracy = cross_entropy_loss_and_accuracy(logits, tokens, valid)
    print(f"Cross Entropy Loss (Function 1): {ce_loss}, Accuracy: {ce_accuracy}")

    # Computation using logits and class indices (tokens) as the representation of true labels.
    fused_ce_loss, fused_accuracy = fused_cross_entropy_loss_and_accuracy(
        logits, tokens, valid
    )
    print(
        f"Fused Cross Entropy Loss (Function 2): {fused_ce_loss}, Accuracy: {fused_accuracy}"
    )

    # Computation using logits and one-hot encoded labels to calculate the loss for each example.
    softmax_ce_loss = softmax_cross_entropy(logits, labels)
    print(f"Softmax Cross Entropy Loss (Function 2): {softmax_ce_loss.mean()}")

    z_loss = 0.0001
    total_loss, total_z_loss = cross_entropy_with_logits(logits, labels, z_loss)
    print(total_loss, total_z_loss)
    print(
        f"Cross Entropy with Logits Loss: {total_loss.mean()}, Z-loss: {total_z_loss.mean()}"
    )

    decoder_loss_weights = jnp.ones_like(tokens, dtype=jnp.float32)

    batch = {
        "decoder_target_tokens": tokens,
        "decoder_loss_weights": decoder_loss_weights,
    }

    label_smoothing = 0.0
    z_loss = 0.0001

    loss_normalizing_factor = SpecialLossNormalizingFactor.NUM_REAL_TARGET_TOKENS
    lnf, weights = get_loss_normalizing_factor_and_weights(
        loss_normalizing_factor, batch
    )

    (
        total_loss,
        total_z_loss,
        weight_sum,
        ce_accuracy,
    ) = compute_weighted_cross_entropy_and_accuracy(
        logits=logits,
        targets=batch["decoder_target_tokens"],
        weights=weights,
        label_smoothing=label_smoothing,
        z_loss=z_loss,
        loss_normalizing_factor=lnf,
    )

    print(
        f"Weighted Cross Entropy Loss: {total_loss}, Z-loss: {total_z_loss}, Weight Sum: {weight_sum}, Accuracy: {ce_accuracy}"
    )

    assert np.isclose(ce_loss, total_loss, rtol=1e-2, atol=1e-2), "Cross Entropy Losses do not match"
    assert np.isclose(ce_accuracy, fused_accuracy, rtol=1e-2, atol=1e-2), "Accuracies do not match"


if __name__ == "__main__":
    test_compute_weighted_cross_entropy_ignore_token_minus100()
    test_compute_weighted_cross_entropy_1()
    test_multiple()



from typing import Optional
import optax


def get_rmsprop_with_cosine_scheduler(
        steps: int,
        learning_rate: float = 5e-5,
        decay: float = 0.9,
        initial_scale: float = 0.,
        momentum: Optional[float] = None,
        nesterov: bool = False,
        eps: float = 1e-8,
        weight_decay: float = 1e-1,
        gradient_accumulation_steps: int = 1,

):

    """
    The get_rmsprop_with_cosine_scheduler function returns a tuple of the optimizer and scheduler.
    The optimizer is composed of several transformations:
        1) scale_by_rms - scales the gradient by RMS (root-mean-square) values, which are calculated using an
         exponential moving average with decay rate `decay` and initial value `initial_scale`. The epsilon
         parameter prevents division by zero.
        2) scale_by_schedule - scales the gradient by a schedule, in this case cosine decay with initial value
        `learning rate` and number of steps to complete one cycle equal to total number of training steps.


    :param steps: int: Set the number of steps in the cosine decay schedule
    :param learning_rate: float: Set the initial learning rate
    :param decay: float: Control the decay rate of the running average
    :param initial_scale: float: Set the initial scale of the rmsprop optimizer
    :param momentum: Optional[float]: Specify the momentum of the optimizer
    :param nesterov: bool: Determine whether to use the nesterov momentum algorithm
    :param eps: float: Avoid division by zero
    :param weight_decay: float: Add a weight decay to the loss function
    :param gradient_accumulation_steps: int: Accumulate the gradients over multiple steps before updating the weights
    :param : Define the number of steps to be taken before the learning rate is decayed
    :return: The optimizer and the scheduler
    """
    scheduler = optax.cosine_decay_schedule(
        init_value=learning_rate,
        decay_steps=steps
    )

    tx = optax.chain(
        optax.scale_by_rms(
            decay=decay,
            eps=eps,
            initial_scale=initial_scale
        ),
        optax.scale_by_schedule(scheduler),
        optax.scale(-1),
        (
            optax.trace(
                decay=momentum,
                nesterov=nesterov
            )
            if momentum is not None else optax.identity()
        ),
        optax.add_decayed_weights(
            weight_decay=weight_decay
        ),
    )
    if gradient_accumulation_steps > 1:
        tx = optax.MultiSteps(
            tx, gradient_accumulation_steps
        )
    return tx, scheduler


def get_rmsprop_with_linear_scheduler(
        steps: int,
        learning_rate_start: float = 5e-5,
        learning_rate_end: float = 1e-5,
        decay: float = 0.9,
        initial_scale: float = 0.,
        momentum: Optional[float] = None,
        nesterov: bool = False,
        eps: float = 1e-8,
        weight_decay: float = 1e-1,
        gradient_accumulation_steps: int = 1,

):

    """
    The get_rmsprop_with_linear_scheduler function returns a tuple of two objects:
        1. A transformation (tx) that is applied to the gradients before they are used to update the model parameters.
        2. A scheduler object that can be used to retrieve the current learning rate at any point during training.

    :param steps: int: Define how many steps the learning rate will take to transition from learning_rate_start to learning_rate_end
    :param learning_rate_start: float: Set the initial learning rate
    :param learning_rate_end: float: Specify the final learning rate
    :param decay: float: Control the decay rate of the rmsprop algorithm
    :param initial_scale: float: Scale the initial gradient
    :param momentum: Optional[float]: Set the momentum of the optimizer
    :param nesterov: bool: Determine whether to use nesterov momentum or not
    :param eps: float: Prevent division by zero
    :param weight_decay: float: Apply weight decay to the model weights
    :param gradient_accumulation_steps: int: Accumulate the gradients over multiple batches
    :param : Control the learning rate decay
    :return: Optimizer,Scheduler
    """
    scheduler = optax.linear_schedule(
        init_value=learning_rate_start,
        end_value=learning_rate_end,
        transition_steps=steps
    )

    tx = optax.chain(
        optax.scale_by_rms(
            decay=decay,
            eps=eps,
            initial_scale=initial_scale
        ),
        optax.scale_by_schedule(scheduler),
        optax.scale(-1),
        (
            optax.trace(
                decay=momentum,
                nesterov=nesterov
            )
            if momentum is not None else optax.identity()
        ),
        optax.add_decayed_weights(
            weight_decay=weight_decay
        ),
    )
    if gradient_accumulation_steps > 1:
        tx = optax.MultiSteps(
            tx, gradient_accumulation_steps
        )
    return tx, scheduler


def get_rmsprop_with_warmup_linear_scheduler(
        steps: int,
        learning_rate_start: float = 5e-5,
        learning_rate_end: float = 1e-5,
        decay: float = 0.9,
        eps: float = 1e-8,
        initial_scale: float = 0.,
        momentum: Optional[float] = None,
        nesterov: bool = False,
        weight_decay: float = 1e-1,
        gradient_accumulation_steps: int = 1,
        warmup_steps: int = 500
):

    """
    The get_rmsprop_with_warmup_linear_scheduler function returns a tuple of the following:
        1. A JAX optimizer transformation (tx) that performs RMSprop with warmup and linear decay, as well as weight decay.
        2. A JAX schedule object (scheduler_combined) that can be used to plot the learning rate over time.

    :param steps: int: Define the number of steps in the training loop
    :param learning_rate_start: float: Set the learning rate at the start of training
    :param learning_rate_end: float: Set the learning rate at the end of training
    :param decay: float: Control the decay rate of the moving average
    :param eps: float: Avoid division by zero
    :param initial_scale: float: Set the initial scale of the rmsprop optimizer
    :param momentum: Optional[float]: Set the momentum of the optimizer
    :param nesterov: bool: Determine whether to use the nesterov momentum algorithm
    :param weight_decay: float: Add a weight decay to the loss function
    :param gradient_accumulation_steps: int: Accumulate the gradients over multiple batches
    :param warmup_steps: int: Set the number of steps to warm up the learning rate
    :return: Optimizer,Scheduler
    """
    scheduler_warmup = optax.linear_schedule(
        init_value=5e-8,
        end_value=learning_rate_start,
        transition_steps=warmup_steps
    )
    scheduler_decay = optax.linear_schedule(
        init_value=learning_rate_start,
        end_value=learning_rate_end,
        transition_steps=steps - warmup_steps
    )

    scheduler_combined = optax.join_schedules(
        schedules=[scheduler_warmup, scheduler_decay],
        boundaries=[warmup_steps]
    )

    tx = optax.chain(
        optax.scale_by_rms(
            decay=decay,
            eps=eps,
            initial_scale=initial_scale
        ),
        optax.scale_by_schedule(scheduler_combined),
        optax.scale(-1),
        (
            optax.trace(
                decay=momentum,
                nesterov=nesterov
            )
            if momentum is not None else optax.identity()
        ),
        optax.add_decayed_weights(
            weight_decay=weight_decay
        ),
    )
    if gradient_accumulation_steps > 1:
        tx = optax.MultiSteps(
            tx, gradient_accumulation_steps
        )
    return tx, scheduler_combined


def get_rmsprop_with_warm_up_cosine_scheduler(
        steps: int,
        learning_rate: float = 5e-5,
        learning_rate_end: float = 1e-5,
        decay: float = 0.9,
        initial_scale: float = 0.,
        momentum: Optional[float] = None,
        nesterov: bool = False,
        eps: float = 1e-8,
        weight_decay: float = 1e-1,
        exponent: float = 1.0,
        gradient_accumulation_steps: int = 1,
        warmup_steps: int = 500,
):
    """
    The get_rmsprop_with_warm_up_cosine_scheduler function returns a tuple of two objects:
        1. A transformation (tx) that is applied to the gradients before they are used to update the parameters.
        2. A scheduler object that can be used to get the current learning rate at any given step in training.

    :param steps: int: Define the number of steps in the warm up phase
    :param learning_rate: float: Set the learning rate of the optimizer
    :param learning_rate_end: float: Set the final learning rate of the optimizer after decay
    :param decay: float: Control the decay rate of the rmsprop algorithm
    :param initial_scale: float: Scale the initial gradient
    :param momentum: Optional[float]: Define the momentum of the optimizer
    :param nesterov: bool: Indicate whether to use nesterov momentum
    :param eps: float: Avoid division by zero
    :param weight_decay: float: Add a weight decay to the loss function
    :param exponent: float: Control the shape of the cosine curve
    :param gradient_accumulation_steps: int: Accumulate gradients over multiple batches
    :param warmup_steps: int: Number of steps of the linear warmup
    :param : Control the learning rate
    :return: Optimizer,Scheduler
    """
    scheduler = optax.warmup_cosine_decay_schedule(
        init_value=0.5e-7,
        peak_value=learning_rate,
        warmup_steps=warmup_steps,
        decay_steps=steps,
        end_value=learning_rate_end,
        exponent=exponent
    )

    tx = optax.chain(
        optax.scale_by_rms(
            decay=decay,
            eps=eps,
            initial_scale=initial_scale
        ),
        optax.scale_by_schedule(scheduler),
        optax.scale(-1),
        (
            optax.trace(
                decay=momentum,
                nesterov=nesterov
            )
            if momentum is not None else optax.identity()
        ),
        optax.add_decayed_weights(
            weight_decay=weight_decay
        ),
    )
    if gradient_accumulation_steps > 1:
        tx = optax.MultiSteps(
            tx, gradient_accumulation_steps
        )
    return tx, scheduler



import flax.core
import jax.tree_util
import optax
import termcolor
from jax import numpy as jnp

from functools import partial
import warnings

from .implicit_array import (
    ImplicitArray,
    use_implicit_args,
    tree_map_with_implicit,
    default_handler,
    primitive_handler,
    ArrayValue,
    aux_field,
    EmptyNode,
    materialize_nested,
    freeze_keys,
    freeze_subtrees,
)

import jax

LORA_FREEZE = 0
LORA_FULL = -1

from dataclasses import dataclass
from typing import Optional, Callable, Any, Union


@dataclass
class LoraWeight(ImplicitArray):
    w: ArrayValue  # M x N
    a: ArrayValue  # k x N
    b: ArrayValue  # M x k

    alpha: float = aux_field(default=1.)

    def __post_init__(self):
        """
        The __post_init__ function is called after the __init__ function.
        It allows us to check that the shapes of our parameters are correct, and if not, raise an error.
        
        :param self: Represent the instance of the class
        :return: The output of the super()
        
        """
        super().__post_init__()
        assert self.a.shape[-2] == self.b.shape[-1], "A and B Array should be like a[K,N] B[M,K]"
        assert self.w.shape[-2] == self.b.shape[-2], "A and B Array should be like W[M,N] B[M,K]"
        assert self.w.shape[-1] == self.a.shape[-1], "A and B Array should be like W[M,N] A[K,N]"

    def materialize(self):
        """
        The materialize function is used to create a new matrix from the parameters of the factorization.
        
        :param self: Access the attributes and methods of a class
        :return: The materialized vector
        
        """
        return (self.w + self.get_scale() * self.b @ self.a).astype(self.w.dtype)

    def get_scale(self):
        """
        The get_scale function returns the scale of the model.
        The scale is defined as alpha / number of columns in b.
        
        
        :param self: Represent the instance of the class
        :return: The scale of the model
        
        """
        return self.alpha / self.b.shape[1]


def _check_dot_dimension_numbers(dimension_numbers):
    (lhs_contract, rhs_contract), (lhs_batch, rhs_batch) = dimension_numbers
    if lhs_batch or rhs_batch:
        warnings.warn("Lorax does not support batched matmuls")
        return False
    if len(lhs_contract) != 1 or len(rhs_contract) != 1:
        warnings.warn("Lorax only supports matmul")
        return False
    return True


@primitive_handler("dot_general")
def handle_dot_lhs(
        primitive,
        lora: LoraWeight,
        rhs: ArrayValue,
        *,
        dimension_numbers,
        **kwargs
):
    """
    The handle_dot_lhs function is a JAX primitive that allows us to perform
    matrix multiplication on LoraWeights. It does this by first performing the matrix
    multiplication on the underlying weight tensor, and then adding in a second term
    that accounts for the fact that we are multiplying two low-rank matrices together.
    
    :param primitive: Determine which function to use
    :param lora: LoraWeight: Pass the loraweight object to the function
    :param rhs: ArrayValue: Pass the right hand side of the dot product
    :param dimension_numbers: Determine which dimensions are being contracted
    :param kwargs: Pass the dimension_numbers to handle_dot_lhs
    :return: The result of the dot product between
    
    """
    if not _check_dot_dimension_numbers(dimension_numbers):
        return NotImplemented

    if isinstance(rhs, LoraWeight):
        rhs = rhs.materialize()
        warnings.warn("Encountered product of two LoraWeights. Materializing the rhs")

    op = partial(jax.lax.dot_general, **kwargs)

    lhs_contract, = dimension_numbers[0][0]

    first, second = (lora.a, lora.b) if lhs_contract == 1 else (lora.b, lora.a)

    first *= lora.get_scale()

    orig = op(lora.w, rhs, dimension_numbers=dimension_numbers)
    lora_product = op(first, rhs, dimension_numbers=dimension_numbers)

    second_dimension_numbers = ((lhs_contract,), (0,)), dimension_numbers[1]

    lora_product = op(second, lora_product, dimension_numbers=second_dimension_numbers)

    return (orig + lora_product).astype(orig.dtype)


@primitive_handler("dot_general")
def handle_dot_rhs(
        primitive,
        lhs: jax.Array,
        lora: LoraWeight,
        *,
        dimension_numbers,
        **kwargs
):
    """
    The handle_dot_rhs function is a partial application of the jax.lax.dot_general function,
    which takes in two arrays and returns their dot product (or matrix multiplication). The 
    handle_dot_rhs function is used to handle the case where a LoraWeight object appears on the right-hand side of an equation. 
    The handle_dot_rhs function takes in three arguments: primitive, lhs, and lora (the LoraWeight object). It then checks that 
    the dimension numbers are correct for this operation using _check_dimension numbers(). If they are not correct it will return NotIm
    
    :param primitive: Identify the function that is being called
    :param lhs: jax.Array: Store the left hand side of the dot product
    :param lora: LoraWeight: Pass the loraweight object to the function
    :param dimension_numbers: Specify the dimensions of the input arrays
    :param kwargs: Pass the dimension_numbers argument to handle_dot_rhs
    :return: The output of the dot product with a loraweight
    
    """
    if not _check_dot_dimension_numbers(dimension_numbers):
        return NotImplemented
    op = partial(jax.lax.dot_general, **kwargs)

    rhs_contract, = dimension_numbers[0][1]
    first, second = (lora.a, lora.b) if rhs_contract == 1 else (lora.b, lora.a)

    first *= lora.get_scale()

    orig = op(lhs, lora.w, dimension_numbers=dimension_numbers)
    lora_product = op(lhs, first, dimension_numbers=dimension_numbers)

    second_dimension_numbers = ((lhs.ndim - 1), (rhs_contract,)), dimension_numbers[1]

    lora_product = op(lora_product, second, dimension_numbers=second_dimension_numbers)

    return (orig + lora_product).astype(orig.dtype)


@primitive_handler("conv_general_dilated")
def handle_conv(
        primitive,
        inp: ArrayValue,
        lora: LoraWeight,
        *,
        dimension_numbers,
        **params
):
    """
    The handle_conv function is a helper function that allows us to use LoraWeight objects as inputs to convolutions.
    
    :param primitive: Identify the function that is being called
    :param inp: ArrayValue: Specify the input to the convolution
    :param lora: LoraWeight: Pass the loraweight object into the function
    :param *: Pass in the dimension_numbers parameter
    :param dimension_numbers: Specify the convolution
    :param params: Pass in the dimension_numbers parameter
    :return: The result of the convolution
    
    """
    if isinstance(inp, LoraWeight):
        warnings.warn("Using a LoraWeight as input to a convolution is not supported, so it will be materialized.")
        inp = inp.materialize()

    if not dimension_numbers.rhs_spec[:1] != (
            len(dimension_numbers.rhs_spec) - 1,
            len(dimension_numbers.rhs_spec) - 2,
    ):
        raise ValueError("Lorax only supports convolutions with shape (..., in_features, out_features)")

    params = {**params, "dimension_numbers": dimension_numbers}
    op = partial(jax.lax.conv_general_dilated, **params)
    orig = op(inp, lora.w)

    lora_product = op(inp, lora.b)

    params["window_strides"] = (1,) * (len(dimension_numbers.rhs_spec) - 2)
    params["padding"] = "VALID"
    lora_product = jax.lax.conv_general_dilated(
        lora_product,
        lora.a * lora.get_scale(),
        **params
    )

    return (orig + lora_product).astype(orig.dtype)


@primitive_handler("gather")
def handle_gather(
        primitive,
        lora: LoraWeight,
        indices: jax.Array,
        *,
        dimension_numbers,
        slice_sizes,
        **params
):
    """
    The handle_gather function is a JAX primitive handler that allows us to
    perform the gather operation on LoraWeight objects. This function is called by
    JAX when it encounters a gather operation in the computation graph. The function
    takes as input:
    
    :param primitive: Identify the operation
    :param lora: LoraWeight: Pass the loraweight object to the function
    :param indices: jax.Array: Select the rows of the weight matrix
    :param dimension_numbers: Specify the dimension numbers of
    :param slice_sizes: Specify the size of each slice
    :param params: Pass the dimension_numbers parameter to the gather function
    :return: A new loraweight
    
    """
    if dimension_numbers.offset_dims != (len(indices.shape) - 1,):
        return NotImplemented

    lora_dim = lora.b.shape[-1]

    if slice_sizes != (1, lora.a.shape[1]):
        return NotImplemented

    params = {**params, "dimension_numbers": dimension_numbers}

    orig = jax.lax.gather(lora.w, indices, slice_sizes=slice_sizes, **params)

    new_slice_sizes = (1, lora_dim)

    lora_product = jax.lax.gather(lora.b, indices, slice_sizes=new_slice_sizes, **params)
    lora_product = lora_product @ (lora.a * lora.get_scale())

    return (orig + lora_product).astype(orig.dtype)


@primitive_handler("transpose")
def eval_lora_transpose(
        primitive,
        arg: LoraWeight,
        *,
        permutation
):
    """
    The eval_lora_transpose function is used to transpose a LoraWeight object.
    
    :param primitive: Determine which function to use
    :param arg: LoraWeight: Specify the type of input that is expected
    :param *: Indicate that the permutation parameter is a keyword-only argument
    :param permutation: Specify the permutation of the weights
    :return: A loraweight object with the same
    
    """
    if not len(arg.shape) == 2 and permutation == (1, 0):
        return NotImplemented

    return LoraWeight(
        w=arg.w.T,
        a=arg.b.T,
        b=arg.a.T,
        alpha=arg.alpha,
    )


@primitive_handler("convert_element_type")
def eval_lora_convert_element_type(primitive, arg: LoraWeight, **params):
    result = jax.tree_map(
        partial(default_handler, primitive, **params),
        arg
    )
    result.dtype = params["new_dtype"]
    return result


def split_lora_params(params, lora_spec):
    """
    Map params to a pytree in which all `LoraWeight.w` values and all params marked with
    LORA_FREEZE are replaced with EmptyNode. This is useful for checkpointing just
    the trainable params.
    """

    def node_mapper(node, spec_val):
        if not isinstance(node, LoraWeight):
            return node if spec_val != LORA_FREEZE else EmptyNode
        children, aux = node.tree_flatten_with_keys()
        idx = next(i for i, (key, _) in enumerate(children) if key == "w")
        children[idx] = ("w", EmptyNode)

        return LoraWeight.tree_unflatten(aux, [c for _, c in children])

    return tree_map_with_implicit(node_mapper, params, lora_spec)


@dataclass
class XRapTureConfig:
    lora_dim: int
    fully_fine_tune_parameters: Optional[list[str]] = None
    lora_fine_tune_parameters: Optional[list[str]] = None
    tune_vectors: bool = True
    verbose: bool = True
    dtype: jnp.dtype = jnp.float32


@dataclass
class XRapTureModule:
    lora_opt_state: Union[jax.tree_util.PyTreeDef, dict]
    lora_parameters: Union[jax.tree_util.PyTreeDef, dict]
    lora_module: Union[Any , flax.linen.Module]
    lora_specs: Union[jax.tree_util.PyTreeDef, dict]
    lora_tx: optax.GradientTransformation


class XRapTure:
    def __init__(
            self,
            config: XRapTureConfig
    ):
        self.config = config

    @staticmethod
    def merge_parameters(
            lora_parameters,
            destructive=True,
    ):

        """    
        The merge_parameters function is used to convert a LoraWeight into an array.
        
        :param lora_parameters: Pass in the parameters of the model
        :param destructive: Determine whether to delete the original parameters or not
        :param : Determine if the function is destructive or not
        :return: The parameters of the model
        
        """

        def _ensure_delete(val):
            if not isinstance(val, jax.Array) or val.is_deleted():
                return
            try:
                val.device_buffer.delete()
            except ValueError:
                val.device_buffers.delete()

        materialize = jax.jit(materialize_nested, donate_argnums=0 if destructive else ())

        def map_fn(param):
            if isinstance(param, LoraWeight):
                result = materialize(param)
                if destructive:
                    jax.tree_map(_ensure_delete, param)
                return result
            return param

        return tree_map_with_implicit(map_fn, lora_parameters)

    def base_decision_function(
            self,
            path: list[jax.tree_util.DictKey],
            params: Optional[Union[dict, jax.tree_util.PyTreeDef]] = None
    ):

        """    
        The base_decision_function function is used to determine which parameters are frozen,
        which are fine-tuned with LoRA, and which are fully fine-tuned. The function takes in a path
        to the parameter (e.g., &quot;model/dense_layer/kernel&quot;) and returns an integer indicating how 
        the parameter should be treated:
        
        :param self: Refer to the object itself
        :param path: list[jax.tree_util.DictKey]: Determine the path of the parameter in question
        :param params: dict | jax.tree_util.PyTreeDef | None: Specify the parameters of the model
        :return: The following:
        
        """
        if self.config.fully_fine_tune_parameters is not None:
            for param_name in self.config.fully_fine_tune_parameters:
                if jax.tree_util.DictKey(key=param_name) in path:
                    if self.config.verbose:
                        print(
                            termcolor.colored(
                                f"Parameter"
                                f" {'/'.join(str(n.key) for n in path)} "
                                f"Selected for Fully Fine-Tune.",
                                color="cyan",
                                force_color=True
                            )
                        )
                    return LORA_FULL

        if self.config.lora_fine_tune_parameters is not None:
            for param_name in self.config.lora_fine_tune_parameters:
                if jax.tree_util.DictKey(key=param_name) in path:
                    if self.config.verbose:
                        print(
                            termcolor.colored(
                                f"Parameter"
                                f" {'/'.join(str(n.key) for n in path)} "
                                f"Selected for LoRA Fine-Tune with {self.config.lora_dim} dimensions.",
                                color="cyan",
                                force_color=True
                            )
                        )
                    return self.config.lora_dim

        return LORA_FREEZE

    def make_lora_specs(
            self,
            parameters: Union[dict, flax.core.FrozenDict],
            decision_fn: Optional[Callable] = None,
            tune_vectors: bool = False,
    ):

        """    
        The make_lora_specs function is used to create a dictionary of LORA specs for the parameters
        of a model. The function takes in two arguments:
        
        :param self: Allow the function to access other attributes and methods of the class
        :param parameters: dict | flax.core.FrozenDict: Specify the parameters to be tuned
        :param decision_fn: Optional[Callable]: Decide whether to freeze or unfreeze a parameter
        :param tune_vectors: bool: Determine if the vectors should be tuned or not
        :param : Decide whether to freeze the parameter or not
        :return: A dictionary of the same shape as the input parameters,
        
        """
        decision_fn = decision_fn if decision_fn is not None else self.base_decision_function

        if decision_fn is None:
            def decision_fn(*args):
                return LORA_FREEZE

        def full_fn(path, arr):
            if len(arr.shape) < 2:
                return LORA_FULL if tune_vectors else LORA_FREEZE

            value = decision_fn(path, arr)
            return value

        return jax.tree_util.tree_map_with_path(full_fn, parameters, is_leaf=None)

    @staticmethod
    def init_lora_parameters(
            param_tree,
            lora_spec,
            dtype: jnp.dtype = jnp.float32,
            rng: jax.random.PRNGKey = jax.random.PRNGKey(0),
            stddev: float = 0.01,
            alpha: float = 1.,
            is_leaf: bool = None
    ):

        """
        The init_lora_parameters function takes in a parameter tree, the lora_spec, and some other parameters.
        It then iterates through the parameter tree using jax.tree_util.tree_map_with_path to get each path and value of 
        the parameter tree (which is just a nested dictionary). It then checks if that value is either LORA_FREEZE or
        LORA_FULL (these are constants defined above). If it's one of those two values, it returns the original parameter as-is; 
        otherwise it creates a new LoraWeight object with random values for b
        
        :param param_tree: Specify the parameters of a neural network
        :param lora_spec: Determine how many parameters to tune
        :param dtype: jnp.dtype: Specify the data type of the parameters
        :param rng: jax.random.PRNGKey: Generate random numbers
        :param stddev: float: Initialize the weights of the network
        :param alpha: float: Control the amount of regularization
        :param is_leaf: bool: Specify whether a node is a leaf or not
        :return: A tree of loraweight objects
        
        """

        def iter_keys(key):
            while True:
                key, out_key = jax.random.split(key)
                yield out_key

        key_it = iter_keys(rng)

        def get_param(path, param, spec_val):
            if spec_val in (LORA_FREEZE, LORA_FULL):
                return param

            if len(param.shape) == 1:
                raise ValueError(
                    f"Vectors must either be frozen or fully tuned, but got "
                    f"lora_spec value {lora_spec} for param with path {path}"
                )

            if len(param.shape) == 2:
                b_dim, a_dim = param.shape

                b = jnp.zeros((b_dim, spec_val), dtype=dtype)
                a = jax.random.normal(next(key_it), (spec_val, a_dim), dtype=dtype) * stddev
                return LoraWeight(w=param, a=a, b=b, alpha=alpha)

            # conv case
            *window_shape, in_channels, out_channels = param.shape

            a = jnp.zeros((
                *(1 for _ in range(len(window_shape))),
                spec_val,
                out_channels
            ), dtype=param.dtype)
            b = jax.random.normal(
                rng, (
                    *window_shape,
                    in_channels,
                    spec_val
                ), dtype=param.dtype
            ) * stddev
            return LoraWeight(
                param,
                a,
                b,
                alpha=alpha
            )

        return jax.tree_util.tree_map_with_path(
            get_param,
            param_tree,
            lora_spec,
            is_leaf=is_leaf
        )

    @staticmethod
    def wrap_tx(
            tx: optax.GradientTransformation,
            lora_spec,
            scalar_frozen_grads=False
    ):

        """    
        The wrap_tx function takes a gradient transformation and wraps it in two
        freeze transformations. The first freezes all parameters that are marked as
        LORA_FREEZE, which is the default for LoraWeight objects. The second freezes
        the weights of all LoraWeight objects, regardless of their freeze status.
        
        :param tx: optax.GradientTransformation: Pass in the optimizer
        :param lora_spec: Specify which parameters we want to freeze
        :param scalar_frozen_grads: Determine whether to use scalar zeros or array zeros
        :return: A transformed version of the optimizer
        
        """
        full_freeze_labels = jax.tree_map(
            lambda x: "freeze" if x == LORA_FREEZE else "train",
            lora_spec
        )
        optimizer_with_full_freeze = freeze_subtrees(
            tx,
            full_freeze_labels,
            use_scalar_zeros=scalar_frozen_grads
        )

        return freeze_keys(
            optimizer_with_full_freeze,
            LoraWeight,
            "w",
            use_scalar_zeros=scalar_frozen_grads
        )

    def apply_lora(
            self,
            module: Union[Any, flax.linen.Module],
            parameters: Union[dict, flax.core.FrozenDict],
            tx: optax.GradientTransformation,
            decision_fn: Optional[Callable] = None,
            tune_vectors: bool = False,
            rng: jax.random.PRNGKey = jax.random.PRNGKey(0),
            stddev: float = 0.01,
            alpha: float = 1.,
            is_leaf: bool = None
    ) -> XRapTureModule:

        """
        The apply_lora function is a wrapper for the XRapTureModule class.
        It takes in a module, parameters, and an optimizer (tx) and returns an instance of the XRapTureModule class.
        The apply_lora function also allows you to specify whether you want to tune vectors as well as
        whether you want to use a decision function when tuning your parameters. The default behavior is that
        vectors are tuned using LORA while scalars are tuned using SGD.

        :param self: Access the attributes of the class
        :param module: Any | flax.linen.Module: Specify the model that is being trained
        :param parameters: dict | flax.core.FrozenDict: Define the parameters of the model
        :param tx: optax.GradientTransformation: Specify the optimizer
        :param decision_fn: Optional[Callable]: Decide whether to apply lora to a parameter or not
        :param tune_vectors: bool: Determine whether to tune the vectors or not
        :param rng: jax.random.PRNGKey: Set the random seed for the initialisation of parameters
        :param stddev: float: Set the standard deviation of the initial weights
        :param alpha: float: Control the variance of the gaussian distribution used to initialize
        :param is_leaf: bool: Determine if the node is a leaf or not
        :return: A XRaptureModule object
        """
        lora_spec = self.make_lora_specs(
            parameters=parameters,
            tune_vectors=tune_vectors,
            decision_fn=decision_fn
        )

        lora_parameters = self.init_lora_parameters(
            param_tree=parameters,
            dtype=self.config.dtype,
            lora_spec=lora_spec,
            rng=rng,
            stddev=stddev,
            alpha=alpha,
            is_leaf=is_leaf
        )
        tx = self.wrap_tx(
            tx=tx,
            lora_spec=lora_spec
        )
        opt_state = tx.init(lora_parameters)
        lora_model = use_implicit_args(
            module
        )

        return XRapTureModule(
            lora_opt_state=opt_state,
            lora_module=lora_model,
            lora_specs=lora_spec,
            lora_parameters=lora_parameters,
            lora_tx=tx
        )



# Modified Implementation of Flash attention
from __future__ import annotations

import math
import os

import functools
from typing import Any, Optional

import jax
from jax import lax
from jax.experimental import pallas as pl
import jax.numpy as jnp
import numpy as np

DEFAULT_MASK_VALUE = -0.7 * float(np.finfo(np.dtype("float32")).max)


def flash_attention_forward_kernel(
        q_ref,
        k_ref,
        v_ref,  # Input arrays
        b_ref: jax.Array | None,  # bias
        o_ref: Any,  # Output
        *residual_refs: Any,  # Residual outputs
        num_heads: int,
        sm_scale: float,
        block_q: int,
        block_d: int,
        block_k: int,
):
    seq_len = q_ref.shape[0]
    start_q = pl.program_id(0)
    if sm_scale is None:
        sm_scale = 1 / math.sqrt(q_ref.shape[-1])
    m_i = jnp.zeros(block_q, dtype=jnp.float32) - float("inf")
    l_i = jnp.zeros(block_q, dtype=jnp.float32)
    o = jnp.zeros((block_q, block_d), dtype=jnp.float32)

    curr_q_slice = pl.dslice(start_q * block_q, block_q)
    q = pl.load(q_ref, (curr_q_slice, pl.dslice(None)))

    def body(start_k, carry):
        o_prev, m_prev, l_prev = carry
        curr_k_slice = pl.dslice(start_k * block_k, block_k)

        k = pl.load(k_ref, (curr_k_slice, slice(None)))

        qk = pl.dot(q, k.T)
        if sm_scale != 1.:
            qk *= sm_scale

        if b_ref is not None:
            b = pl.load(b_ref, (curr_q_slice, curr_k_slice))
            qk = jnp.add(b, qk, )

        m_curr = qk.max(axis=-1)
        m_next = jnp.maximum(m_prev, m_curr)
        correction = jnp.exp(m_prev - m_next)
        l_prev_corr = correction * l_prev
        s_curr = jnp.exp(qk - m_next[:, None])
        l_curr = s_curr.sum(axis=-1)
        l_next = l_prev_corr + l_curr
        l_next_rcp = 1. / l_next
        s_curr = s_curr * l_next_rcp[:, None]
        o_prev_corr = (l_prev_corr * l_next_rcp)[:, None] * o_prev
        v = pl.load(v_ref, (curr_k_slice, pl.dslice(block_d)))
        o_curr = pl.dot(s_curr.astype(v.dtype), v)

        o_next = o_prev_corr + o_curr
        return o_next, m_next, l_next

    upper_bound = pl.cdiv(seq_len, block_k)
    o, m_i, l_i = lax.fori_loop(0, upper_bound, body, (o, m_i, l_i))

    if residual_refs:
        l_ref, m_ref = residual_refs
        pl.store(l_ref, (curr_q_slice,), l_i)
        pl.store(m_ref, (curr_q_slice,), m_i)
    # Write output to dram.
    o = o.astype(o_ref.dtype)
    pl.store(o_ref, (curr_q_slice, pl.dslice(None)), o)


@functools.partial(
    jax.custom_vjp, nondiff_argnums=[4, 5, 6, 7, 8, 9, 10, 11, 12]
)
@functools.partial(
    jax.jit,
    static_argnames=[
        "sm_scale",
        "block_q",
        "block_k",
        "backward_pass_impl",
        "num_warps",
        "num_stages",
        "grid",
        "interpret",
        "debug",
    ],
)
def flash_attention(
        query,
        key,
        value,
        bias: Optional[jnp.ndarray] = None,
        sm_scale: Optional[float] = None,
        block_q: int = 128,
        block_k: int = 128,
        backward_pass_impl: str = "triton",
        num_warps: Optional[int] = None,
        num_stages: int = 2,
        grid: Optional[tuple[int, ...]] = None,
        interpret: Optional[bool] = None,
        debug: bool = False,
):
    del backward_pass_impl

    batch_size, seq_len, num_heads, head_dim = query.shape
    if sm_scale is None:
        sm_scale = 1 / math.sqrt(head_dim)
    if interpret is None:
        interpret = not (seq_len / 16).is_integer() or jax.lib.xla_bridge.get_backend().platform == "cpu"

    block_q = min(block_q, seq_len)
    block_k = min(block_k, seq_len)
    # Heuristics.
    grid_ = grid
    if grid_ is None:
        grid_ = (pl.cdiv(seq_len, block_q), batch_size, num_heads)

    num_warps_ = num_warps
    if num_warps_ is None:
        num_warps_ = 4 if head_dim <= 64 else 8
    kernel = functools.partial(
        flash_attention_forward_kernel,
        num_heads=num_heads,
        sm_scale=sm_scale,
        block_q=block_q,
        block_k=block_k,
        block_d=head_dim,
    )

    in_specs = [
        pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
        pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
        pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
        None if bias is None else pl.BlockSpec(lambda _, j, k: (j, 0, 0, 0), (None, None, seq_len, seq_len))
    ]
    out_shape = jax.ShapeDtypeStruct(shape=query.shape, dtype=query.dtype)
    return pl.pallas_call(
        kernel,
        grid=grid_,
        in_specs=in_specs,
        out_specs=pl.BlockSpec(
            lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
        ),
        compiler_params=dict(
            triton=dict(num_warps=num_warps_, num_stages=num_stages)
        ),
        out_shape=out_shape,
        debug=debug,
        interpret=interpret,
        name="flash_attention_forward",
    )(query, key, value, bias)


def _flash_attention_forward(
        query,
        key,
        value,
        bias: jax.Array | None,
        sm_scale: float,
        block_q: int,
        block_k: int,
        backward_pass_impl: str,
        num_warps: int | None,
        num_stages: int,
        grid: Any,
        interpret: bool,
        debug: bool,
):
    del backward_pass_impl
    batch_size, seq_len, num_heads, head_dim = query.shape
    if sm_scale is None:
        sm_scale = 1 / math.sqrt(head_dim)
    if interpret is None:
        interpret = not (seq_len / 16).is_integer() or jax.lib.xla_bridge.get_backend().platform == "cpu"
    block_q = min(block_q, seq_len)
    block_k = min(block_k, seq_len)
    # Heuristics.
    grid_ = grid
    if grid_ is None:
        grid_ = (pl.cdiv(seq_len, block_q), batch_size, num_heads)

    num_warps_ = num_warps
    if num_warps_ is None:
        num_warps_ = 4 if head_dim <= 64 else 8
    kernel = functools.partial(
        flash_attention_forward_kernel,
        num_heads=num_heads,
        sm_scale=sm_scale,
        block_q=block_q,
        block_k=block_k,
        block_d=head_dim
    )
    out_shape = [
        jax.ShapeDtypeStruct(shape=query.shape, dtype=query.dtype),  # out
        jax.ShapeDtypeStruct(shape=(batch_size, num_heads, seq_len), dtype=jnp.float32),
        jax.ShapeDtypeStruct(shape=(batch_size, num_heads, seq_len), dtype=jnp.float32)
    ]
    in_specs = [
        pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
        pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
        pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
        None if bias is None else pl.BlockSpec(lambda _, j, k: (j, 0, 0, 0), (None, None, seq_len, seq_len))
    ]
    out_specs = [
        pl.BlockSpec(
            lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
        ),
        pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
        pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
    ]
    out, l, m = pl.pallas_call(
        kernel,
        grid=grid_,
        in_specs=in_specs,
        out_specs=out_specs,  # type:ignore
        compiler_params=dict(
            triton=dict(num_warps=num_warps_, num_stages=num_stages)
        ),
        out_shape=out_shape,
        debug=debug,
        interpret=interpret,
        name="flash_attention_forward",
    )(query, key, value, bias)
    return out, (query, key, value, bias, out, l, m)


def _preprocess_backward_kernel(
        out_ref,
        d_out_ref,
        l_ref,
        new_d_out_ref,
        delta_ref, *,
        block_q: int
):
    pid_m = pl.program_id(0)

    off_m = pl.ds(pid_m * block_q, block_q)
    # load
    o = pl.load(out_ref, (off_m, slice(None))).astype(jnp.float32)
    do = pl.load(d_out_ref, (off_m, slice(None))).astype(jnp.float32)
    de_num = pl.load(l_ref, (off_m,)).astype(jnp.float32)
    # compute
    do = do / de_num[:, None]
    delta = jnp.sum(o * do, axis=1)
    # write-back
    pl.store(new_d_out_ref, (off_m, slice(None)), do.astype(new_d_out_ref.dtype))
    pl.store(delta_ref, (off_m,), delta.astype(delta_ref.dtype))


@jax.named_scope("preprocess_backward")
def _preprocess_backward(
        out,
        do,
        l,
        block_q: int,
        debug: bool,
        interpret: bool
):
    batch_size, seq_len, num_heads, head_dim = out.shape
    out_shape = [
        jax.ShapeDtypeStruct(do.shape, do.dtype),
        jax.ShapeDtypeStruct(l.shape, l.dtype),
    ]
    out_specs = [
        pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
        pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
    ]
    do_scaled, delta = pl.pallas_call(
        functools.partial(_preprocess_backward_kernel, block_q=block_q),
        grid=(pl.cdiv(seq_len, block_q), batch_size, num_heads),
        in_specs=[
            pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
            pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
            pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
        ],
        out_specs=out_specs,  # type:ignore
        compiler_params=dict(
            triton=dict(num_warps=4, num_stages=3)
        ),
        out_shape=out_shape,
        debug=debug,
        interpret=interpret,
        name="flash_attention_preprocess_backward")(out, do, l)
    return do_scaled, delta


def flash_attention_backward_kernel(
        # Inputs
        q_ref,
        k_ref,
        v_ref,
        b_ref: jax.Array | None,
        out_ref,
        do_scaled_ref,
        l_ref,
        m_ref,
        delta_ref,
        _,
        # Outputs
        dq_ref,
        dk_ref,
        dv_ref,
        *,
        sm_scale: float,
        block_q: int,
        block_d: int,
        block_k: int,
):
    del out_ref, l_ref  # Not needed
    seq_len = q_ref.shape[0]

    def outer_loop(start_k, _):

        dv = jnp.zeros([block_k, block_d], dtype=jnp.float32)
        dk = jnp.zeros([block_k, block_d], dtype=jnp.float32)
        k = pl.load(k_ref, (pl.ds(start_k * block_k, block_k), slice(None)))
        v = pl.load(v_ref, (pl.ds(start_k * block_k, block_k), slice(None)))

        def inner_loop(start_q, carry):
            dv, dk = carry
            q = pl.load(q_ref, (pl.ds(start_q * block_q, block_q), slice(None)))
            qk = pl.dot(q, k.T)
            qk = qk.astype(q_ref.dtype)
            qk = qk.astype(jnp.float32)
            if sm_scale != 1.0:
                qk *= sm_scale
            if b_ref is not None:
                b = pl.load(b_ref, (pl.ds(start_q * block_q, block_q), pl.ds(start_k * block_k, block_k)))
                qk = jnp.add(b, qk)

            m = pl.load(m_ref, (pl.ds(start_q * block_q, block_q),))
            p = jnp.exp(qk - m[:, None])
            do = pl.load(do_scaled_ref, (pl.ds(start_q * block_q, block_q), slice(None)))
            dv = dv + pl.dot(p.astype(do.dtype).T, do)
            di = pl.load(delta_ref, (pl.ds(start_q * block_q, block_q),))
            dp = jnp.zeros((block_q, block_k), dtype=jnp.float32) - di[:, None]
            dp = dp + pl.dot(do, v.T)
            ds = p * dp
            if sm_scale != 1.0:
                ds = ds * sm_scale
            dk = dk + pl.dot(ds.astype(q_ref.dtype).T, q)
            dq = pl.load(dq_ref, (pl.ds(start_q * block_q, block_q), slice(None)), eviction_policy="evict_last")
            dq = dq + pl.dot(ds.astype(k.dtype), k).astype(dq.dtype)
            pl.store(dq_ref, (pl.ds(start_q * block_q, block_q), slice(None)), dq, eviction_policy="evict_last")
            return dv, dk

        dv, dk = lax.fori_loop(0, pl.cdiv(seq_len, block_q), inner_loop, (dv, dk))
        pl.store(dv_ref, (pl.ds(start_k * block_k, block_k), slice(None)), dv.astype(dv_ref.dtype))
        pl.store(dk_ref, (pl.ds(start_k * block_k, block_k), slice(None)), dk.astype(dk_ref.dtype))

    lax.fori_loop(0, pl.cdiv(seq_len, block_k), outer_loop, None)


@functools.partial(jax.jit, static_argnames=["sm_scale"])
def _flash_attention_reference(
        q,
        k,
        v,
        b: Optional[jax.Array] = None,
        sm_scale=1.0,
):
    logits = jnp.einsum("bqhc,bkhc->bhqk", q, k).astype(jnp.float32)
    logits = b + logits
    weights = jax.nn.softmax(logits * sm_scale).astype(q.dtype)
    return jnp.einsum("bhqk,bkhc->bqhc", weights, v)


def _flash_attention_backward(
        sm_scale: float,
        block_q: int,
        block_k: int,
        backward_pass_impl: str,
        num_warps: int | None,
        num_stages: int,
        grid: Any,
        interpret: bool,
        debug: bool,
        res,
        do
):
    del num_warps, num_stages, grid
    q, k, v, b, out, l, m = res

    if sm_scale is None:
        sm_scale = 1 / math.sqrt(q.shape[-1])
    if backward_pass_impl == "xla":
        return jax.vjp(
            functools.partial(_flash_attention_reference, sm_scale=sm_scale),
            q,
            k,
            v,
            b,
        )[1](do)
    elif backward_pass_impl == "triton":
        batch_size, seq_len, num_heads, head_dim = q.shape
        block_q = min(block_q, seq_len)
        block_k = min(block_k, seq_len)
        do_scaled, delta = _preprocess_backward(out, do, l, block_q, debug, interpret)
        dq = jnp.zeros(q.shape, jnp.float32)
        out_shapes = [
            jax.ShapeDtypeStruct(dq.shape, dq.dtype),
            jax.ShapeDtypeStruct(k.shape, k.dtype),
            jax.ShapeDtypeStruct(v.shape, v.dtype),
        ]

        in_specs = [
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(lambda j, k: (j, k, 0), (None, None, seq_len)),
            pl.BlockSpec(lambda j, k: (j, k, 0), (None, None, seq_len)),
            pl.BlockSpec(lambda j, k: (j, k, 0), (None, None, seq_len)),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
        ]
        if b is None:
            in_specs.insert(3, None)  # type: ignore[arg-type]
            input_output_aliases = {8: 0}
        else:
            in_specs.insert(3, pl.BlockSpec(lambda j, k: (j, 0, 0, 0), (None, None, seq_len, seq_len)))
            input_output_aliases = {9: 0}
        grid = (batch_size, num_heads)
        num_warps = 8
        out_specs = [
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
        ]
        dq, dk, dv = pl.pallas_call(
            functools.partial(
                flash_attention_backward_kernel,
                block_q=block_q,
                block_d=head_dim,
                block_k=block_k,
                sm_scale=sm_scale,
            ),
            out_specs=out_specs,  # type:ignore
            grid=grid,
            out_shape=out_shapes,
            in_specs=in_specs,
            name="flash_attention_backward",
            debug=debug,
            interpret=interpret,
            compiler_params=dict(
                triton=dict(
                    num_warps=num_warps, num_stages=1
                )
            ),
            input_output_aliases=input_output_aliases,
        )(q, k, v, b, out, do_scaled, l, m, delta, dq)
    else:
        raise ValueError(f"Invalid backward pass implementation: {backward_pass_impl}")
    return dq.astype(q.dtype), dk, dv, None


flash_attention.defvjp(_flash_attention_forward, _flash_attention_backward)



from .linen import (
    Linear as Linear,
    LinearBitKernel as LinearBitKernel,
    quantize as quantize,
    de_quantize as de_quantize,
    quantize_parameters as quantize_parameters,
    de_quantize_params as de_quantize_params,
    Conv as Conv,
    Embed as Embed,
    promote_dtype as promote_dtype,
    ConvTranspose as ConvTranspose,
    GroupNorm as GroupNorm,
    BatchNorm as BatchNorm,
    LayerNorm as LayerNorm,
    RMSNorm as RMSNorm,
    WeightNorm as WeightNorm,
    InstanceNorm as InstanceNorm,
    SpectralNorm as SpectralNorm,
    Module as Module,
    ConvLocal as ConvLocal,
    compact as compact,
    initializers as initializers,

)

from flax.linen import *
from flax.linen import Dropout

__all__ = (
    "Linear",
    "LinearBitKernel",
    "quantize",
    "de_quantize",
    "quantize_parameters",
    "de_quantize_params",
    "Conv",
    "Embed",
    "promote_dtype",
    "ConvTranspose",
    "GroupNorm",
    "BatchNorm",
    "LayerNorm",
    "RMSNorm",
    "WeightNorm",
    "InstanceNorm",
    "SpectralNorm",
    "Module",
    "ConvLocal",
    "compact",
    "initializers",
    "GeGLU",
    "PReLU",
    "celu",
    "elu",
    "gelu",
    "glu",
    "hard_sigmoid",
    "hard_silu",
    "hard_swish",
    "hard_tanh",
    "leaky_relu",
    "log_sigmoid",
    "log_softmax",
    "logsumexp",
    "normalize",
    "one_hot",
    "relu6",
    "relu",
    "selu",
    "sigmoid",
    "silu",
    "soft_sign",
    "softmax",
    "softplus",
    "standardize",
    "swish",
    "tanh",
    "add_metadata_axis",
    "checkpoint",
    "cond",
    "custom_vjp",
    "jit",
    "jvp",
    "map_variables",
    "named_call",
    "remat_scan",
    "remat",
    "scan",
    "switch",
    "vjp",
    "grad",
    "value_and_grad",
    "vmap",
    "while_loop",
    "LogicallyPartitioned",
    "get_logical_axis_rules",
    "logical_axis_rules",
    "logical_to_mesh",
    "logical_to_mesh_axes",
    "logical_to_mesh_sharding",
    "set_logical_axis_rules",
    "with_logical_constraint",
    "with_logical_partitioning",
    "Dropout"
)
